from derivaml_test import TestDerivaML
from deriva_ml import MLVocab as vc, Workflow, ExecutionConfiguration, DatasetSpec
from deriva_ml.demo_catalog import (
    reset_demo_catalog,
    populate_demo_catalog,
    create_demo_datasets,
)


class TestExecution(TestDerivaML):
    def test_execution_no_download(self):
        reset_demo_catalog(self.ml_instance, self.domain_schema)
        self.ml_instance.add_term(
            vc.workflow_type,
            "Manual Workflow",
            description="Initial setup of Model File",
        )
        self.ml_instance.add_term(
            vc.execution_asset_type,
            "API_Model",
            description="Model for our API workflow",
        )
        self.ml_instance.add_term(
            vc.workflow_type,
            "ML Demo",
            description="A ML Workflow that uses Deriva ML API",
        )

        api_workflow = Workflow(
            name="Manual Workflow",
            url="https://github.com/informatics-isi-edu/deriva-ml/blob/main/tests/test_execution.py",
            workflow_type="Manual Workflow",
            description="A manual operation",
        )

        manual_execution = self.ml_instance.create_execution(
            ExecutionConfiguration(
                description="Sample Execution", workflow=api_workflow
            )
        )
        manual_execution.upload_execution_outputs()

    def test_execution_download(self):
        populate_demo_catalog(self.ml_instance, self.domain_schema)
        create_demo_datasets(self.ml_instance)
        exec_config = execution_test(self.ml_instance)
        exec = self.ml_instance.create_execution(exec_config)


def execution_test(ml_instance):
    training_dataset_rid = [
        ds["RID"]
        for ds in ml_instance.find_datasets()
        if "Training" in ds["Dataset_Type"]
    ][0]
    testing_dataset_rid = [
        ds["RID"]
        for ds in ml_instance.find_datasets()
        if "Testing" in ds["Dataset_Type"]
    ][0]

    nested_dataset_rid = [
        ds["RID"]
        for ds in ml_instance.find_datasets()
        if "Partitioned" in ds["Dataset_Type"]
    ][0]

    ml_instance.add_term(
        vc.workflow_type, "Manual Workflow", description="Initial setup of Model File"
    )
    ml_instance.add_term(
        vc.execution_asset_type, "API_Model", description="Model for our API workflow"
    )
    ml_instance.add_term(
        vc.workflow_type, "ML Demo", description="A ML Workflow that uses Deriva ML API"
    )
    api_workflow = Workflow(
        name="Manual Workflow",
        url="https://github.com/informatics-isi-edu/deriva-ml/blob/main/docs/Notebooks/DerivaML%20Execution.ipynb",
        workflow_type="Manual Workflow",
        description="A manual operation",
    )

    manual_execution = ml_instance.create_execution(
        ExecutionConfiguration(description="Sample Execution", workflow=api_workflow)
    )

    # Now lets create model configuration for our program.
    model_file = manual_execution.execution_asset_path("API_Model") / "modelfile.txt"
    with open(model_file, "w") as fp:
        fp.write(f"My model")

    # Now upload the file and retrieve the RID of the new asset from the returned results.
    uploaded_assets = manual_execution.upload_execution_outputs()

    training_model_rid = uploaded_assets["API_Model/modelfile.txt"].result["RID"]
    api_workflow = Workflow(
        name="ML Demo",
        url="https://github.com/informatics-isi-edu/deriva-ml/blob/main/pyproject.toml",
        workflow_type="ML Demo",
        description="A workflow that uses Deriva ML",
    )

    config = ExecutionConfiguration(
        datasets=[
            DatasetSpec(
                rid=nested_dataset_rid,
                version=ml_instance.dataset_version(nested_dataset_rid),
            ),
            DatasetSpec(
                rid=testing_dataset_rid,
                version=ml_instance.dataset_version(testing_dataset_rid),
            ),
        ],
        assets=[training_model_rid],
        description="Sample Execution",
        workflow=api_workflow,
    )
    return config



from derivaml_test import TestDerivaML

from deriva_ml import MLVocab, Workflow, ExecutionConfiguration
from tempfile import TemporaryDirectory
from random import random
from deriva_ml.demo_catalog import (
    reset_demo_catalog,
)


class TestUpload(TestDerivaML):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def test_upload_directory(self):
        reset_demo_catalog(self.ml_instance, self.domain_schema)
        self.ml_instance.create_asset("FooBar")
        with TemporaryDirectory() as tmpdir:
            asset_dir = self.ml_instance.asset_dir("FooBar", prefix=tmpdir)
            for s in range(2):
                asset_file = asset_dir.create_file(f"test_{s}.txt", metadata={})
                with open(asset_file, "w+") as f:
                    f.write(f"Hello there {random()}\n")
            self.ml_instance.upload_assets(asset_dir)
        assets = list(
            self.ml_instance.catalog.getPathBuilder()
            .schemas[self.domain_schema]
            .tables["FooBar"]
            .entities()
            .fetch()
        )
        self.assertEqual(len(assets), 2)

    def test_upload_directory_metadata(self):
        reset_demo_catalog(self.ml_instance, self.domain_schema)
        subject_path = (
            self.ml_instance.catalog.getPathBuilder()
            .schemas[self.domain_schema]
            .tables["Subject"]
        )
        ss = list(subject_path.insert([{"Name": f"Thing{t + 1}"} for t in range(2)]))
        with TemporaryDirectory() as tmpdir:
            image_dir = self.ml_instance.asset_dir("Image", prefix=tmpdir)
            for s in ss:
                image_file = image_dir.create_file(
                    f"test_{s['RID']}.txt", {"Subject": s["RID"]}
                )
                with open(image_file, "w+") as f:
                    f.write(f"Hello there {random()}\n")
            self.ml_instance.upload_assets(image_dir)
        assets = list(
            self.ml_instance.catalog.getPathBuilder()
            .schemas[self.domain_schema]
            .tables["Image"]
            .entities()
            .fetch()
        )
        self.assertIn(assets[0]["Subject"], [s["RID"] for s in ss])
        self.assertEqual(len(assets), 2)

    def test_upload_execution_outputs(self):
        reset_demo_catalog(self.ml_instance, self.domain_schema)
        self.ml_instance.add_term(
            MLVocab.workflow_type,
            "Manual Workflow",
            description="Initial setup of Model File",
        )
        self.ml_instance.add_term(
            MLVocab.execution_asset_type,
            "API_Model",
            description="Model for our API workflow",
        )

        api_workflow = Workflow(
            name="Manual Workflow",
            url="https://github.com/informatics-isi-edu/deriva-ml/blob/main/tests/test_upload.py",
            workflow_type="Manual Workflow",
            description="A manual operation",
        )

        manual_execution = self.ml_instance.create_execution(
            ExecutionConfiguration(
                description="Sample Execution", workflow=api_workflow
            )
        )

        # Now let us create model configuration for our program.
        model_file = (
            manual_execution.execution_asset_path("API_Model") / "modelfile.txt"
        )
        with open(model_file, "w") as fp:
            fp.write(f"My model")

        # Now upload the file and retrieve the RID of the new asset from the returned results.
        uploaded_assets = manual_execution.upload_execution_outputs()
        path = self.ml_instance.catalog.getPathBuilder().schemas["deriva-ml"]
        self.assertEqual(1, len(list(path.Execution_Asset.entities().fetch())))
        execution_metadata = list(path.Execution_Metadata.entities().fetch())
        print([m["Filename"] for m in execution_metadata])
        self.assertEqual(2, len(execution_metadata))



import sys
import argparse
from deriva.core import get_credential, ErmrestCatalog
from deriva.core.ermrest_model import Model

from deriva_ml.schema_setup.annotations import generate_annotation


def alter_table_annotation(
    model_root: Model, schema_name: str, table_name: str, annotation: dict
):
    table = model_root.schemas[schema_name].tables[table_name]
    table.alter(annotations=annotation)
    model_root.apply()


def alter_schema_annotation(model_root: Model, schema_name: str, annotation: dict):
    schema = model_root.schemas[schema_name]
    schema.alter(annotations=annotation)
    model_root.apply()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--hostname", type=str, required=True)
    parser.add_argument("--schema_name", type=str, required=True)
    parser.add_argument("--catalog_id", type=str, required=True)
    args = parser.parse_args()
    credentials = get_credential(args.hostname)
    catalog = ErmrestCatalog("https", args.hostname, args.catalog_id, credentials)
    model = catalog.getCatalogModel()
    annotations = generate_annotation(model)
    alter_table_annotation(
        model, args.schema_name, "Workflow", annotations["workflow_annotation"]
    )
    alter_table_annotation(
        model, args.schema_name, "Execution", annotations["execution_annotation"]
    )
    alter_table_annotation(
        model,
        args.schema_name,
        "Execution_Metadata",
        annotations["execution_metadata_annotation"],
    )
    alter_table_annotation(
        model,
        args.schema_name,
        "Execution_Asset",
        annotations["execution_asset_annotation"],
    )
    alter_schema_annotation(model, args.schema_name, annotations["schema_annotation"])


if __name__ == "__main__":
    sys.exit(main())



import sys

from deriva.core import DerivaServer, get_credential
from deriva.core.ermrest_model import builtin_types, Schema, Table, Column, ForeignKey
from deriva.chisel import Model, Schema, Table, Column, ForeignKey
from deriva_ml.schema_setup.annotation_temp import generate_annotation
import argparse


def create_schema_if_not_exist(model, schema_name, schema_comment=None):
    if schema_name not in model.schemas:
        schema = model.create_schema(Schema.define(schema_name, schema_comment))
        return schema
    else:
        schema = model.schemas[schema_name]
        return schema


def create_table_if_not_exist(schema, table_name, create_spec):
    if table_name not in schema.tables:
        table = schema.create_table(create_spec)
        return table
    else:
        table = schema.tables[table_name]
        return table


def define_table_workflow(workflow_annotation: dict):
    table_def = Table.define(
        'Workflow',
        column_defs=[
            Column.define('Name', builtin_types.text),
            Column.define('Description', builtin_types.markdown),
            Column.define('URL', builtin_types.ermrest_uri),
            Column.define('Checksum', builtin_types.text),
            Column.define('Version', builtin_types.text)
        ],
        fkey_defs=[
            ForeignKey.define(
                ['RCB'],
                'public', 'ERMrest_Client',
                ['ID']
            )
        ],
        annotations=workflow_annotation
    )
    return table_def


def define_table_execution(execution_annotation: dict):
    table_def = Table.define(
        'Execution',
        column_defs=[
            Column.define('Description', builtin_types.markdown),
            Column.define('Duration', builtin_types.text),
            Column.define('Status', builtin_types.text),
            Column.define('Status_Detail', builtin_types.text)
        ],
        fkey_defs=[
            ForeignKey.define(
                ['RCB'],
                'public', 'ERMrest_Client',
                ['ID']
            )
        ],
        annotations=execution_annotation
    )
    return table_def


def define_asset_execution_metadata(schema: str, execution_metadata_annotation: dict):
    table_def = Table.define_asset(
        sname=schema,
        tname='Execution_Metadata',
        hatrac_template='/hatrac/metadata/{{MD5}}.{{Filename}}',
        fkey_defs=[
            ForeignKey.define(
                ['RCB'],
                'public', 'ERMrest_Client',
                ['ID']
            )
        ],
        annotations=execution_metadata_annotation
    )
    return table_def


def define_asset_execution_assets(schema: str, execution_assets_annotation: dict):
    table_def = Table.define_asset(
        sname=schema,
        tname='Execution_Assets',
        hatrac_template='/hatrac/execution_assets/{{MD5}}.{{Filename}}',
        fkey_defs=[
            ForeignKey.define(
                ['RCB'],
                'public', 'ERMrest_Client',
                ['ID']
            )
        ],
        annotations=execution_assets_annotation
    )
    return table_def


def setup_ml_workflow(model, schema_name, catalog_id):
    curie_template = catalog_id+':{RID}'
    schema = create_schema_if_not_exist(model, schema_name)
    # get annotations
    annotations = generate_annotation(catalog_id, schema_name)
    # Workflow
    workflow_table = create_table_if_not_exist(schema, 'Workflow',
                                               define_table_workflow(annotations["workflow_annotation"]))
    table_def_workflow_type_vocab = Table.define_vocabulary(
        tname='Workflow_Type', curie_template=curie_template
    )
    workflow_type_table = schema.create_table(table_def_workflow_type_vocab)
    workflow_table.add_reference(workflow_type_table)

    # Execution
    execution_table = create_table_if_not_exist(schema, 'Execution',
                                                define_table_execution(annotations["execution_annotation"]))
    execution_table.add_reference(workflow_table)
    # dataset_table = create_table_if_not_exist(schema, 'Dataset', define_table_dataset(schema))
    # association_dataset_execution = schema.create_association(dataset_table, execution_table)

    # Execution Metadata
    execution_metadata_table = create_table_if_not_exist(schema, 'Execution_Metadata',
                                                         define_asset_execution_metadata(schema,
                                                                                         annotations["execution_metadata_annotation"]))
    execution_metadata_table.add_reference(execution_table)
    table_def_metadata_type_vocab = Table.define_vocabulary(tname='Execution_Metadata_Type',
                                                            curie_template=curie_template)
    metadata_type_table = schema.create_table(table_def_metadata_type_vocab)
    execution_metadata_table.add_reference(metadata_type_table)

    # Execution Asset
    execution_assets_table = create_table_if_not_exist(schema, 'Execution_Assets',
                                                       define_asset_execution_assets(schema,
                                                                                     annotations["execution_assets_annotation"]))
    association_execution_execution_asset = schema.create_association(execution_assets_table, execution_table)

    table_def_execution_product_type_vocab = Table.define_vocabulary(
        tname='Execution_Asset_Type', curie_template=curie_template
    )
    execution_asset_type_table = schema.create_table(table_def_execution_product_type_vocab)
    execution_assets_table.add_reference(execution_asset_type_table)
    # image_table = create_table_if_not_exist(schema, 'Image', define_asset_image(schema))
    # association_image_execution_asset = schema.create_association(execution_assets_table, image_table)


def main():
    scheme = 'https'
    parser = argparse.ArgumentParser()
    parser.add_argument('--hostname', type=str, required=True)
    parser.add_argument('--schema_name', type=str, required=True)
    parser.add_argument('--catalog_id', type=str, required=True)
    args = parser.parse_args()
    credentials = get_credential(args.hostname)
    server = DerivaServer(scheme, args.hostname, credentials)
    model = Model.from_catalog(server.connect_ermrest(args.catalog_id))
    setup_ml_workflow(model, args.schema_name, args.catalog_id)


if __name__ == "__main__":
    sys.exit(main())



"""
`deriva_ml_base.py` is the core module for the Deriva ML project.  This module implements the DerivaML class, which is
the primary interface to the Deriva based catalogs.  The module also implements the Feature and Vocabulary functions
in the DerivaML.

DerivaML and its associated classes all depend on a catalog that implements a `deriva-ml` schema with tables and
relationships that follow a specific data model.

"""

import getpass
import logging
from datetime import datetime
from itertools import chain
from pathlib import Path
from typing import Optional, Any, Iterable, TYPE_CHECKING
from deriva.core import (
    ErmrestCatalog,
    get_credential,
    urlquote,
    DEFAULT_SESSION_CONFIG,
)
from deriva.core.datapath import DataPathException, _ResultSet as ResultSet
from deriva.core.datapath import _CatalogWrapper
from deriva.core.ermrest_catalog import ResolveRidResult
from deriva.core.ermrest_model import FindAssociationResult, Key, Table
from deriva.core.hatrac_store import HatracStore
from pydantic import validate_call, ConfigDict

from .execution_configuration import ExecutionConfiguration
from .feature import Feature, FeatureRecord
from .dataset import Dataset
from .deriva_model import DerivaModel
from .upload import (
    table_path,
    execution_rids,
    execution_metadata_dir,
    upload_directory,
    UploadAssetDirectory,
)
from .deriva_definitions import ColumnDefinition
from .deriva_definitions import ExecMetadataVocab
from .deriva_definitions import (
    RID,
    Status,
    FileUploadState,
    DerivaMLException,
    ML_SCHEMA,
    VocabularyTerm,
)

if TYPE_CHECKING:
    from .execution import Execution


class DerivaML(Dataset):
    """Base class for ML operations on a Deriva catalog.

    This class is intended to be used as a base class on which more domain specific interfaces are built.

    Attributes:
        host_name: Hostname of the Deriva server.
        catalog_id: Catalog ID. Either and identifier, or a catalog name.
        domain_schema: Schema name for domain specific tables and relationships.
        model: ERMRest model for the catalog
    """

    def __init__(
        self,
        hostname: str,
        catalog_id: str | int,
        domain_schema: str = None,
        project_name: str = None,
        cache_dir: Optional[str] = None,
        working_dir: Optional[str] = None,
        model_version: str = "1",
        ml_schema: str = ML_SCHEMA,
        logging_level=logging.INFO,
    ):
        """Create and initialize a DerivaML instance.

        This method will connect to a catalog, and initialize local configuration for the ML execution.
        This class is intended to be used as a base class on which domain-specific interfaces are built.

        Args:
            hostname: Hostname of the Deriva server.
            catalog_id: Catalog ID. Either and identifier, or a catalog name.
            domain_schema: Schema name for domain specific tables and relationships.
            cache_dir: Directory path for caching data downloaded from the Deriva server as bdbag.
            working_dir: Directory path for storing data used by or generated by any computations.
            model_version: A string that indicates the version model.  Typically passed in via
        """
        self.credential = get_credential(hostname)
        self.catalog = ErmrestCatalog(
            "https",
            hostname,
            catalog_id,
            self.credential,
            session_config=self._get_session_config(),
        )
        self.model = DerivaModel(
            self.catalog.getCatalogModel(), domain_schema=domain_schema
        )

        default_workdir = self.__class__.__name__ + "_working"
        self.working_dir = (
            Path(working_dir) / getpass.getuser()
            if working_dir
            else Path.home() / "deriva-ml"
        ) / default_workdir
        self.working_dir.mkdir(parents=True, exist_ok=True)
        self.cache_dir = (
            Path(cache_dir) if cache_dir else Path.home() / "deriva-ml" / "cache"
        )
        self.cache_dir.mkdir(parents=True, exist_ok=True)

        # Initialize dataset class.
        super().__init__(self.model, self.cache_dir)

        self.host_name = hostname
        self.catalog_id = catalog_id
        self.ml_schema = ml_schema
        self.version = model_version
        self.configuration = None

        self.domain_schema = self.model.domain_schema
        self.project_name = project_name or self.domain_schema

        self.start_time = datetime.now()
        self.status = Status.pending.value
        self._logger = logging.getLogger("deriva_ml")
        self._logger.setLevel(logging_level)

        logging.basicConfig(
            level=logging_level,
            format="%(asctime)s - %(name)s.%(levelname)s - %(message)s",
        )

        # Set logging level for dervia library
        deriva_logger = logging.getLogger("deriva")
        deriva_logger.setLevel(logging_level)

        if "dirty" in self.version:
            logging.info(
                f"Loading dirty model.  Consider commiting and tagging: {self.version}"
            )

    @staticmethod
    def _get_session_config():
        """ """
        session_config = DEFAULT_SESSION_CONFIG.copy()
        session_config.update(
            {
                # our PUT/POST to ermrest is idempotent
                "allow_retry_on_all_methods": True,
                # do more retries before aborting
                "retry_read": 8,
                "retry_connect": 5,
                # increase delay factor * 2**(n-1) for Nth retry
                "retry_backoff_factor": 5,
            }
        )
        return session_config

    @property
    def pathBuilder(self) -> _CatalogWrapper:
        """Get a new instance of a pathBuilder object."""
        return self.catalog.getPathBuilder()

    @property
    def domain_path(self):
        """Get a new instance of a pathBuilder object to the domain schema"""

        return self.pathBuilder.schemas[self.domain_schema]

    def table_path(self, table: str | Table) -> Path:
        """Return a local file path in which to place a CSV to add values to a table on upload.

        Args:
          table: return:
          table: str | Table:

        Returns:
            Path to a CSV file in which to add values to a table on upload.
        """
        return table_path(
            self.working_dir,
            schema=self.domain_schema,
            table=self.model.get_table(table).name,
        )

    def asset_dir(
        self, table: str | Table, prefix: str | Path = None
    ) -> UploadAssetDirectory:
        """Return a local file path in which to place a files for an asset table.  T

        Args:
            table: Location of where to place files.  Defaults to execution_assets_path.
            prefix: Root path to asset directory.

        Returns:
            Path to the directory in which asset files should be placed.
        """
        table = self.model.get_table(table)
        if not self.model.is_asset(table):
            raise DerivaMLException(f"The table {table} is not an asset table.")

        prefix = Path(prefix) if prefix else self.working_dir
        return UploadAssetDirectory(
            model=self.model,
            prefix=prefix,
            schema=table.schema.name,
            table=table.name,
        )

    def download_dir(self, cached: bool = False) -> Path:
        """Location where downloaded files are placed.

        Args:
          cached: bool:  (Default value = False)

        Returns:

        """
        return self.cache_dir if cached else self.working_dir

    def chaise_url(self, table: RID | Table) -> str:
        """Return a Chaise URL to the specified table.

        Args:
            table: Table or RID to be visited
            table: str | Table:

        Returns:
            URL to the table in Chaise format.
        """
        try:
            table = self.model.get_table(table)
            uri = self.catalog.get_server_uri().replace(
                "ermrest/catalog/", "chaise/recordset/#"
            )
        except DerivaMLException:
            # Perhaps we have a RID....
            uri = self.cite(table)
        return f"{uri}/{urlquote(table.schema.name)}:{urlquote(table.name)}"

    def cite(self, entity: dict | str) -> str:
        """Return a citation URL for the provided entity.

        Args:
            entity: A dict that contains the column values for a specific entity or a RID.

        Returns:
            The URI for the provided entity.

        Raises:
            DerivaMLException: if provided RID does not exist.
        """
        if entity.startswith(f"https://{self.host_name}/id/{self.catalog_id}/"):
            # Already got a citation...
            return entity
        try:
            self.resolve_rid(
                rid := entity if isinstance(entity, str) else entity["RID"]
            )
            return f"https://{self.host_name}/id/{self.catalog_id}/{rid}@{self.catalog.latest_snapshot().snaptime}"
        except KeyError as e:
            raise DerivaMLException(f"Entity {e} does not have RID column")
        except DerivaMLException as _e:
            raise DerivaMLException("Entity RID does not exist")

    def user_list(self) -> list[dict[str, str]]:
        """List of users in the catalog

        Args:

        Returns:
          A list of dictionaries containing user information.

        """
        user_path = self.pathBuilder.public.ERMrest_Client.path
        return [
            {"ID": u["ID"], "Full_Name": u["Full_Name"]}
            for u in user_path.entities().fetch()
        ]

    def resolve_rid(self, rid: RID) -> ResolveRidResult:
        """Return a named tuple with information about the specified RID.

        Args:
            rid: RID of the object of interest

        Returns:
            ResolveRidResult which has information about the specified RID.

        Raises:
          DerivaMLException: if the RID doesn't exist.
        """
        try:
            return self.catalog.resolve_rid(rid, self.model.model)
        except KeyError as _e:
            raise DerivaMLException(f"Invalid RID {rid}")

    def retrieve_rid(self, rid: RID) -> dict[str, Any]:
        """Return a dictionary that represents the values of the specified RID.

        Args:
            rid: RID of the object of interest

        Returns:
          A dictionary that represents the values of the specified RID.

        Raises:
          DerivaMLException: if the RID doesn't exist.
        """
        return self.resolve_rid(rid).datapath.entities().fetch()[0]

    def add_page(self, title: str, content: str) -> None:
        """

        Args:
          title: str:
          content: str:

        Returns:

        """
        self.pathBuilder.www.tables[self.domain_schema].insert(
            [{"Title": title, "Content": content}]
        )

    def create_vocabulary(
        self, vocab_name: str, comment: str = "", schema: Optional[str] = None
    ) -> Table:
        """Create a controlled vocabulary table with the given vocab name.

        Args:
            vocab_name: Name of the controlled vocabulary table.
            comment: Description of the vocabulary table. (Default value = '')
            schema: Schema in which to create the controlled vocabulary table.  Defaults to domain_schema.
            vocab_name: str:

        Returns:
            An ERMRest table object for the newly created vocabulary table.
        """
        schema = schema or self.domain_schema
        return self.model.schemas[schema].create_table(
            Table.define_vocabulary(
                vocab_name, f"{self.project_name}:{{RID}}", comment=comment
            )
        )

    def create_asset(
        self,
        asset_name: str,
        column_defs: Iterable[ColumnDefinition] = None,
        comment: str = "",
        schema: str = None,
    ) -> Table:
        """Create an asset table with the given asset name.

        Args:
            asset_name: Name of the asset table.
            column_defs: Iterable of ColumnDefinition objects to provide additional metadata for asset.
            comment: Description of the asset table. (Default value = '')
            schema: Schema in which to create the asset table.  Defaults to domain_schema.
            asset_name: str:
            schema: str:  (Default value = None)

        Returns:
            Table object for the asset table.
        """
        column_defs = column_defs or []
        schema = schema or self.domain_schema
        asset_table = self.model.schemas[schema].create_table(
            Table.define_asset(
                schema,
                asset_name,
                column_defs=[c.model_dump() for c in column_defs],
                comment=comment,
            )
        )
        return asset_table

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def create_feature(
        self,
        target_table: Table | str,
        feature_name: str,
        terms: list[Table | str] = None,
        assets: list[Table | str] = None,
        metadata: Iterable[ColumnDefinition | Table | Key | str] = None,
        optional: Optional[list[str]] = None,
        comment: str = "",
    ) -> type[FeatureRecord]:
        """Create a new feature that can be associated with a table.

        The feature can associate a controlled vocabulary term, an asset, or any other values with a
        specific instance of an object and  execution.

        Args:
            feature_name: Name of the new feature to be defined
            target_table: table name or object on which the feature is to be associated
            terms: List of controlled vocabulary terms that will be part of the feature value
            assets: List of asset table names or objects that will be part of the feature value
            metadata: List of other value types that are associated with the feature
            optional: List of columns that are optional in the feature
            comment: return: A Feature class that can be used to create instances of the feature.

        Returns:
            A Feature class that can be used to create instances of the feature.

        Raises:
            DerivaException: If the feature cannot be created.
        """

        terms = terms or []
        assets = assets or []
        metadata = metadata or []
        optional = optional or []

        def normalize_metadata(m: Key | Table | ColumnDefinition | str):
            """

            Args:
              m: Key | Table | ColumnDefinition | str:

            Returns:

            """
            if isinstance(m, str):
                return self.model.get_table(m)
            elif isinstance(m, ColumnDefinition):
                return m.model_dump()
            else:
                return m

        # Make sure that the provided assets or terms are actually assets or terms.
        if not all(map(self.model.is_asset, assets)):
            raise DerivaMLException(f"Invalid create_feature asset table.")
        if not all(map(self.model.is_vocabulary, terms)):
            raise DerivaMLException(f"Invalid create_feature asset table.")

        # Get references to the necessary tables and make sure that the
        # provided feature name exists.
        target_table = self.model.get_table(target_table)
        execution = self.model.schemas[self.ml_schema].tables["Execution"]
        feature_name_table = self.model.schemas[self.ml_schema].tables["Feature_Name"]
        feature_name_term = self.add_term(
            "Feature_Name", feature_name, description=comment
        )
        atable_name = f"Execution_{target_table.name}_{feature_name_term.name}"

        # Now create the association table that implements the feature.
        atable = self.model.schemas[self.domain_schema].create_table(
            target_table.define_association(
                table_name=atable_name,
                associates=[execution, target_table, feature_name_table],
                metadata=[
                    normalize_metadata(m) for m in chain(assets, terms, metadata)
                ],
                comment=comment,
            )
        )
        # Now set optional terms.
        for c in optional:
            atable.columns[c].alter(nullok=True)
        atable.columns["Feature_Name"].alter(default=feature_name_term.name)
        return self.feature_record_class(target_table, feature_name)

    def feature_record_class(
        self, table: str | Table, feature_name: str
    ) -> type[FeatureRecord]:
        """Create a pydantic model for entries into the specified feature table.

        For information on how to
        See the pydantic documentation for more details about the pydantic model.

        Args:
            table: table name or object on which the feature is to be associated
            feature_name: name of the feature to be created
            table: str | Table:
            feature_name: str:

        Returns:
            A Feature class that can be used to create instances of the feature.
        """
        return self.lookup_feature(table, feature_name).feature_record_class()

    def delete_feature(self, table: Table | str, feature_name: str) -> bool:
        """

        Args:
          table: Table | str:
          feature_name: str:

        Returns:
        """
        table = self.model.get_table(table)
        try:
            feature = next(
                f for f in self.find_features(table) if f.feature_name == feature_name
            )
            feature.feature_table.drop()
            return True
        except StopIteration:
            return False

    def lookup_feature(self, table: str | Table, feature_name: str) -> Feature:
        """Lookup the named feature associated with the provided table.

        Args:
            table: param feature_name:
            table: str | Table:
            feature_name: str:

        Returns:
            A Feature class that represents the requested feature.

        Raises:
          DerivaMLException: If the feature cannot be found.
        """
        table = self.model.get_table(table)
        try:
            return [
                f for f in self.find_features(table) if f.feature_name == feature_name
            ][0]
        except IndexError:
            raise DerivaMLException(
                f"Feature {table.name}:{feature_name} doesn't exist."
            )

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def find_features(self, table: Table | str) -> Iterable[Feature]:
        """List the names of the features in the specified table.

        Args:
            table: The table to find features for.
            table: Table | str:

        Returns:
            An iterable of FeatureResult instances that describe the current features in the table.
        """
        table = self.model.get_table(table)

        def is_feature(a: FindAssociationResult) -> bool:
            """

            Args:
              a: FindAssociationResult:

            Returns:

            """
            # return {'Feature_Name', 'Execution'}.issubset({c.name for c in a.table.columns})
            return {
                "Feature_Name",
                "Execution",
                a.self_fkey.foreign_key_columns[0].name,
            }.issubset({c.name for c in a.table.columns})

        return [
            Feature(a, self.model)
            for a in table.find_associations(min_arity=3, max_arity=3, pure=False)
            if is_feature(a)
        ]

    @validate_call
    def add_features(self, features: Iterable[FeatureRecord]) -> int:
        """Add a set of new feature values to the catalog.

        Args:
          features: Iterable[FeatureRecord]:

        Returns:
            Number of attributes added
        """
        features = list(features)
        feature_table = features[0].feature.feature_table
        feature_path = self.pathBuilder.schemas[feature_table.schema.name].tables[
            feature_table.name
        ]
        entries = feature_path.insert(f.model_dump() for f in features)
        return len(entries)

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def list_feature_values(self, table: Table | str, feature_name: str) -> ResultSet:
        """Return a datapath ResultSet containing all values of a feature associated with a table.

        Args:
            table: param feature_name:
            table: Table | str:
            feature_name: str:

        Returns:

        """
        table = self.model.get_table(table)
        feature = self.lookup_feature(table, feature_name)
        pb = self.catalog.getPathBuilder()
        return (
            pb.schemas[feature.feature_table.schema.name]
            .tables[feature.feature_table.name]
            .entities()
            .fetch()
        )

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def add_term(
        self,
        table: str | Table,
        term_name: str,
        description: str,
        synonyms: Optional[list[str]] = None,
        exists_ok: bool = True,
    ) -> VocabularyTerm:
        """Creates a new control vocabulary term in the control vocabulary table.

        Args:

        Args:
            table: The name of the control vocabulary table.
            term_name: The name of the new control vocabulary.
            description: The description of the new control vocabulary.
            synonyms: Optional list of synonyms for the new control vocabulary. Defaults to an empty list.
            exists_ok: Optional flag indicating whether to allow creation if the control vocabulary name
                already exists. Defaults to True.

        Returns:
          The RID of the newly created control vocabulary.

        Raises:
          DerivaException: If the control vocabulary name already exists and exist_ok is False.
        """
        synonyms = synonyms or []
        table = self.model.get_table(table)
        pb = self.catalog.getPathBuilder()
        if not (self.model.is_vocabulary(table)):
            raise DerivaMLException(f"The table {table} is not a controlled vocabulary")

        schema_name = table.schema.name
        table_name = table.name
        try:
            term_id = VocabularyTerm.model_validate(
                pb.schemas[schema_name]
                .tables[table_name]
                .insert(
                    [
                        {
                            "Name": term_name,
                            "Description": description,
                            "Synonyms": synonyms,
                        }
                    ],
                    defaults={"ID", "URI"},
                )[0]
            )
        except DataPathException:
            term_id = self.lookup_term(table, term_name)
            if not exists_ok:
                raise DerivaMLException(f"{term_name} already exists")
            # Check vocabulary
        return term_id

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def lookup_term(self, table: str | Table, term_name: str) -> VocabularyTerm:
        """Given a term name, return the vocabulary record.  Can provide either the term name
         or a synonym for the term.  Generate an exception if the term is not in the vocabulary.

        Args:
            table: The name of the controlled vocabulary table or a ERMRest table object.
            term_name: The name of the term to look up.

        Returns:
          The entry the associated term or synonym.

        Raises:
          DerivaException: If the schema or vocabulary table doesn't exist, or if the term is not
            found in the vocabulary.
        """
        vocab_table = self.model.get_table(table)
        if not self.model.is_vocabulary(vocab_table):
            raise DerivaMLException(f"The table {table} is not a controlled vocabulary")
        schema_name, table_name = vocab_table.schema.name, vocab_table.name
        schema_path = self.catalog.getPathBuilder().schemas[schema_name]
        for term in schema_path.tables[table_name].entities():
            if term_name == term["Name"] or (
                term["Synonyms"] and term_name in term["Synonyms"]
            ):
                return VocabularyTerm.model_validate(term)
        raise DerivaMLException(f"Term {term_name} is not in vocabulary {table_name}")

    def list_vocabulary_terms(self, table: str | Table) -> list[VocabularyTerm]:
        """Return a list of terms that are in a vocabulary table.

        Args:
            table: The name of the controlled vocabulary table or a ERMRest table object.
            table: str | Table:

        Returns:
            The list of terms that are in a vocabulary table.

        Raises:
            DerivaMLException: If the schema or vocabulary table doesn't exist, or if the table is not
                a controlled vocabulary.
        """
        pb = self.catalog.getPathBuilder()
        table = self.model.get_table(table)
        if not (self.model.is_vocabulary(table)):
            raise DerivaMLException(f"The table {table} is not a controlled vocabulary")

        return [
            VocabularyTerm(**v)
            for v in pb.schemas[table.schema.name].tables[table.name].entities().fetch()
        ]

    def download_asset(self, asset_url: str, dest_filename: str) -> Path:
        """Download an asset from a URL and place it in a local directory.

        Args:
            asset_url: URL of the asset.
            dest_filename: Destination filename.

        Returns:
            A  Path object to the downloaded asset.
        """
        hs = HatracStore("https", self.host_name, self.credential)
        hs.get_obj(path=asset_url, destfilename=dest_filename)
        return Path(dest_filename)

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def upload_assets(
        self,
        assets_dir: str | Path | UploadAssetDirectory,
    ) -> dict[Any, FileUploadState] | None:
        """Upload assets from a directory. This routine assumes that the current upload specification includes a
        configuration for the specified directory.  Every asset in the specified directory is uploaded

        Args:
            assets_dir: Directory containing the assets to upload.

        Returns:
            Results of the upload operation.

        Raises:
            DerivaMLException: If there is an issue uploading the assets.
        """

        def path_to_asset(path: str) -> str:
            components = path.split("/")
            return components[
                components.index("asset") + 2
            ]  # Look for asset in the path to find the name

        if isinstance(assets_dir, UploadAssetDirectory):
            assets_dir = assets_dir.path

        if not self.model.is_asset(Path(assets_dir).name):
            raise DerivaMLException("Directory does not have name of an asset table.")
        results = upload_directory(self.model, assets_dir)
        return {path_to_asset(p): r for p, r in results.items()}

    def _update_status(
        self, new_status: Status, status_detail: str, execution_rid: RID
    ):
        """Update the status of an execution in the catalog.

        Args:
            new_status: New status.
            status_detail: Details of the status.
            execution_rid: Resource Identifier (RID) of the execution.
            new_status: Status:
            status_detail: str:
             execution_rid: RID:

        Returns:

        """
        self.status = new_status.value
        self.pathBuilder.schemas[self.ml_schema].Execution.update(
            [
                {
                    "RID": execution_rid,
                    "Status": self.status,
                    "Status_Detail": status_detail,
                }
            ]
        )

    # @validate_call
    def create_execution(self, configuration: ExecutionConfiguration) -> "Execution":
        """Create an execution object

        Given an execution configuration, initialize the local compute environment to prepare for executing an
        ML or analytic routine.  This routine has a number of side effects.

        1. The datasets specified in the configuration are downloaded and placed in the cache-dir. If a version is
        not specified in the configuration, then a new minor version number is created for the dataset and downloaded.

        Args:
            configuration: ExecutionConfiguration:

        Returns:
            An execution object.
        """
        from .execution import Execution

        return Execution(configuration, self)

    # @validate_call
    def restore_execution(self, execution_rid: Optional[RID] = None) -> "Execution":
        from .execution import Execution

        # Find path to execution
        if not execution_rid:
            e_rids = execution_rids(self.working_dir)
            if len(e_rids) != 1:
                raise DerivaMLException(f"Multiple execution RIDs were found {e_rids}.")

            execution_rid = e_rids[0]
        cfile = (
            execution_metadata_dir(
                self.working_dir,
                exec_rid=execution_rid,
                metadata_type=ExecMetadataVocab.execution_config.value,
            )
            / "configuration.json"
        )
        configuration = ExecutionConfiguration.load_configuration(cfile)
        return Execution(configuration, self, reload=execution_rid)



"""
This module provides functions that help structure local directories for uploading to a DerivaML catalog, and
generating an upload specification for those directories.

Here is the directory layout we support:

  deriva-ml/
       execution
           <execution_rid>
               execution-asset
                   <asset_type>
                       file1, file2, ....   <- Need to update execution_asset association table.
               execution-metadata
                   <metadata_type>
               feature
                   <schema>
                       <target_table>
                            <feature_name>
                                   asset
                                       <asset_table>
                                           file1, file2, ...
                           <feature_name>.csv    <- needs to have asset_name column remapped before uploading
            table
               <schema>
                   <record_table>
                      record_table.csv
            asset
               <schema>
                   <asset_table>
                     file1, file2, ....
        asset
            <schema>
                <asset_table>
                    <metadata1>
                        <metadata2>
                            file1, file2, ....

"""

import json
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Optional

import regex as re
from deriva.core import urlquote
from deriva.core.ermrest_model import Table
from deriva.core.hatrac_store import HatracStore
from deriva.core.utils import hash_utils, mime_utils
from deriva.transfer.upload.deriva_upload import GenericUploader
import logging
from pydantic import validate_call, ConfigDict

from deriva_ml.deriva_definitions import (
    RID,
    DerivaMLException,
    FileUploadState,
    UploadState,
)
from deriva_ml.deriva_model import DerivaModel

upload_root_regex = r"(?i)^.*/deriva-ml"

exec_dir_regex = upload_root_regex + r"/execution/(?P<execution_rid>[-\w]+)"
exec_asset_dir_regex = (
    exec_dir_regex + r"/execution-asset/(?P<execution_asset_type>[-\w]+)"
)
exec_asset_regex = (
    exec_asset_dir_regex + r"/(?P<file_name>[-\w]+)[.](?P<file_ext>[a-z0-9]+)$"
)
exec_metadata_dir_regex = (
    exec_dir_regex + r"/execution-metadata/(?P<execution_metadata_type>[-\w]+)"
)
exec_metadata_regex = (
    exec_metadata_dir_regex + r"/(?P<filename>[-\w]+)[.](?P<file_ext>[a-z0-9]*)$"
)
feature_dir_regex = exec_dir_regex + r"/feature"
feature_table_dir_regex = (
    feature_dir_regex
    + r"/(?P<schema>[-\w]+)/(?P<target_table>[-\w]+)/(?P<feature_name>[-\w]+)"
)
feature_value_regex = (
    feature_table_dir_regex + r"/(?P=feature_name)[.](?P<file_ext>[(csv|json)]*)$"
)
feature_asset_dir_regex = feature_table_dir_regex + r"/asset/(?P<asset_table>[-\w]+)"
feature_asset_regex = (
    feature_asset_dir_regex
    + r"/(?P<file_name>[A-Za-z0-9_-]+)[.](?P<file_ext>[a-z0-9]*)$"
)
asset_path_regex = (
    upload_root_regex + r"/asset/(?P<schema>[-\w]+)/(?P<asset_table>[-\w]*)"
)

asset_file_regex = r"(?P<file_name>[-\w]+)[.](?P<file_ext>[a-z0-9]*)$"

table_regex = (
    exec_dir_regex
    + r"/table/(?P<schema>[-\w]+)/(?P<table>[-\w]+)/(?P=table)[.](csv|json)$"
)


def is_execution_metadata_dir(path: Path) -> Optional[re.Match]:
    """Path matches the patten for execution metadata directory."""
    return re.match(exec_metadata_dir_regex + "$", path.as_posix())


def is_execution_asset_dir(path: Path) -> Optional[re.Match]:
    """Path maths the pattern for execution asset directory"""
    return re.match(exec_asset_dir_regex + "$", path.as_posix())


def is_feature_dir(path: Path) -> Optional[re.Match]:
    """Path matches the pattern for where the table for a feature would go."""
    return re.match(feature_table_dir_regex + "$", path.as_posix())


def is_feature_asset_dir(path: Path) -> Optional[re.Match]:
    """Path matches the pattern for feature, asset, returns the feature components."""
    return re.match(feature_asset_dir_regex + "$", path.as_posix())


def upload_root(prefix: Path | str) -> Path:
    """Return the top level directory of where to put files to be uploaded."""
    path = Path(prefix) / "deriva-ml"
    path.mkdir(exist_ok=True, parents=True)
    return path


def asset_root(prefix: Path | str) -> Path:
    """Return the top level directory of where to put asset directories to be uploaded."""
    path = upload_root(prefix) / "asset"
    path.mkdir(exist_ok=True, parents=True)
    return path


def asset_table_dir(prefix: Path, asset_schema, asset_table) -> Path:
    """Return the top level directory of where to asset table will be uploaded."""
    path = asset_root(prefix) / asset_schema / asset_table
    path.mkdir(exist_ok=True, parents=True)
    return path


def execution_rids(prefix: Path | str) -> list[RID]:
    """Return a list of all the execution RIDS that have files waiting to be uploaded."""
    path = upload_root(prefix) / "execution"
    return [d.name for d in path.iterdir()]


def execution_root(prefix: Path | str, exec_rid) -> Path:
    """Path to directory to place execution specific upload files."""
    path = upload_root(prefix) / "execution" / exec_rid
    path.mkdir(exist_ok=True, parents=True)
    return path


def execution_asset_root(prefix: Path | str, exec_rid: str) -> Path:
    """Path to directory into which execution assets should be located"""

    path = execution_root(prefix, exec_rid) / "execution-asset"
    path.mkdir(parents=True, exist_ok=True)
    return path


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def execution_asset_dir(prefix: Path | str, exec_rid: str, asset_type: str) -> Path:
    """Return the path to a directory in which to place execution assets of a specified type are to be uploaded."""
    path = execution_asset_root(prefix, exec_rid) / asset_type
    path.mkdir(parents=True, exist_ok=True)
    return path


def execution_metadata_root(prefix: Path | str, exec_rid: str) -> Path:
    """Path to directory into which execution metadata should be located."""
    path = execution_root(prefix, exec_rid) / "execution-metadata"
    path.mkdir(parents=True, exist_ok=True)
    return path


def execution_metadata_dir(
    prefix: Path | str, exec_rid: str, metadata_type: str
) -> Path:
    """Return the path to a directory in which to place execution metadata of a specified type are to be uploaded."""

    path = execution_metadata_root(prefix, exec_rid) / metadata_type
    path.mkdir(parents=True, exist_ok=True)
    return path


def feature_root(prefix: Path | str, exec_rid: str) -> Path:
    """Return the path to the directory in which features for the specified execution should be placed."""
    path = execution_root(prefix, exec_rid) / "feature"
    path.mkdir(parents=True, exist_ok=True)
    return path


def feature_dir(
    prefix: Path | str, exec_rid: str, schema: str, target_table: str, feature_name: str
) -> Path:
    """Return the path to eht directory in which a named feature for an execution should be placed."""
    path = feature_root(prefix, exec_rid) / schema / target_table / feature_name
    path.mkdir(parents=True, exist_ok=True)
    return path


def feature_value_path(
    prefix: Path | str, exec_rid: str, schema: str, target_table: str, feature_name: str
) -> Path:
    """Return the path to a CSV file in which to place feature values that are to be uploaded.

    Args:
        prefix: Location of upload root directory
        exec_rid: RID of the execution to be associated with this feature.
        schema: Domain schema name
        target_table: Target table name for the feature.
        feature_name: Name of the feature.

    Returns:
        Path to CSV file in which to place feature values
    """
    return (
        feature_dir(prefix, exec_rid, schema, target_table, feature_name)
        / f"{feature_name}.csv"
    )


def feature_asset_dir(
    prefix: Path | str,
    exec_rid: str,
    schema: str,
    target_table: str,
    feature_name: str,
    asset_table: str,
) -> Path:
    """Return the path to a directory in which to place feature assets for a named feature are to be uploaded.

    Args:
        prefix: Location of upload root directory
        exec_rid: RID of the execution for the feature asset
        schema: Domain schema
        target_table: Name of the target table for the feature.
        feature_name: Name of the feature
        asset_table: Name of the asset table for the feature.

    Returns:
        Path to directory in which feature asset files are placed.
    """
    path = (
        feature_dir(prefix, exec_rid, schema, target_table, feature_name)
        / "asset"
        / asset_table
    )

    path.mkdir(parents=True, exist_ok=True)
    return path


def table_path(prefix: Path | str, schema: str, table: str) -> Path:
    """Return the path to a CSV file in which to place table values that are to be uploaded.

    Args:
        prefix: Location of upload root directory
        schema: Domain schema
        table: Name of the table to be uploaded.

    Returns:
        Path to the file in which to place table values that are to be uploaded.
    """
    path = upload_root(prefix) / "table" / schema / table
    path.mkdir(parents=True, exist_ok=True)
    return path / f"{table}.csv"


def asset_table_upload_spec(model, asset_table):
    """Generate a pattern to an asset table that may include additional metadata columns.

    Args:
        model:
        asset_table:

    Returns:

    """
    metadata_columns = model.asset_metadata(asset_table)
    asset_table = model.get_table(asset_table)
    schema = model.get_table(asset_table).schema.name
    metadata_path = "/".join([rf"(?P<{c}>[-\w]+)" for c in metadata_columns])
    asset_path = f"{upload_root_regex}/asset/{schema}/{asset_table.name}/{metadata_path}/{asset_file_regex}"
    asset_table = model.get_table(asset_table)
    schema = model.get_table(asset_table).schema.name
    return {
        # Upload assets into an asset table of an asset table.
        "column_map": {
            "MD5": "{md5}",
            "URL": "{URI}",
            "Length": "{file_size}",
            "Filename": "{file_name}",
        }
        | {c: f"{{{c}}}" for c in metadata_columns},
        "file_pattern": asset_path,  # Sets schema, asset_table, file_name, file_ext
        "target_table": [schema, asset_table.name],
        "checksum_types": ["sha256", "md5"],
        "hatrac_options": {"versioned_urls": True},
        "hatrac_templates": {
            "hatrac_uri": f"/hatrac/{asset_table.name}/{{md5}}.{{file_name}}",
            "content-disposition": "filename*=UTF-8''{file_name}.{file_ext}",
        },
        "record_query_template": "/entity/{target_table}/MD5={{md5}}&Filename={{file_name}}",
    }


def bulk_upload_configuration(model: DerivaModel) -> dict[str, Any]:
    """Return an upload specification for deriva-ml
    Arguments:
        model: Model from which to generate the upload configuration
    """
    asset_tables_with_metadata = [
        asset_table_upload_spec(model=model, asset_table=t)
        for t in model.find_assets()
        if model.asset_metadata(t)
    ]
    return {
        "asset_mappings": [
            {
                # Upload  any files that may have been created by the program execution.  These are  in the
                # Execution_Metadata directory
                "column_map": {
                    "MD5": "{md5}",
                    "URL": "{URI}",
                    "Length": "{file_size}",
                    "Filename": "{file_name}",
                    "Execution_Metadata_Type": "{execution_metadata_type_name}",
                },
                "file_pattern": exec_metadata_regex,
                "target_table": ["deriva-ml", "Execution_Metadata"],
                "checksum_types": ["sha256", "md5"],
                "hatrac_options": {"versioned_urls": True},
                "hatrac_templates": {
                    "hatrac_uri": "/hatrac/execution_metadata/{md5}.{file_name}",
                    "content-disposition": "filename*=UTF-8''{file_name}.{file_ext}",
                },
                "record_query_template": "/entity/{target_table}/MD5={md5}&Filename={file_name}",
                "metadata_query_templates": [
                    "/attribute/deriva-ml:Execution_Metadata_Type/Name={execution_metadata_type}/execution_metadata_type_name:=Name"
                ],
            },
            {
                # Upload the contents of the Execution_Asset directory.
                "column_map": {
                    "MD5": "{md5}",
                    "URL": "{URI}",
                    "Length": "{file_size}",
                    "Filename": "{file_name}",
                    "Execution_Asset_Type": "{execution_asset_type_name}",
                },
                "file_pattern": exec_asset_regex,
                "target_table": ["deriva-ml", "Execution_Asset"],
                "checksum_types": ["sha256", "md5"],
                "hatrac_options": {"versioned_urls": True},
                "hatrac_templates": {
                    "hatrac_uri": "/hatrac/execution_asset/{md5}.{file_name}",
                    "content-disposition": "filename*=UTF-8''{file_name}.{file_ext}",
                },
                "record_query_template": "/entity/{target_table}/MD5={md5}&Filename={file_name}",
                "metadata_query_templates": [
                    "/attribute/deriva-ml:Execution_Asset_Type/Name={execution_asset_type}/execution_asset_type_name:=Name"
                ],
            },
            {
                # Upload the assets for a feature table.
                "column_map": {
                    "MD5": "{md5}",
                    "URL": "{URI}",
                    "Length": "{file_size}",
                    "Filename": "{file_name}",
                },
                "file_pattern": feature_asset_regex,  # Sets target_table, feature_name, asset_table
                "target_table": ["{schema}", "{asset_table}"],
                "checksum_types": ["sha256", "md5"],
                "hatrac_options": {"versioned_urls": True},
                "hatrac_templates": {
                    "hatrac_uri": "/hatrac/{asset_table}/{md5}.{file_name}",
                    "content-disposition": "filename*=UTF-8''{file_name}",
                },
                "record_query_template": "/entity/{target_table}/MD5={md5}&Filename={file_name}",
            },
        ]
        + asset_tables_with_metadata
        + [
            {
                # Upload assets into an asset table of an asset table without any metadata
                "column_map": {
                    "MD5": "{md5}",
                    "URL": "{URI}",
                    "Length": "{file_size}",
                    "Filename": "{file_name}",
                },
                "target_table": [model.domain_schema, "{asset_table}"],
                "file_pattern": asset_path_regex
                + "/"
                + asset_file_regex,  # Sets schema, asset_table, file_name, file_ext
                "checksum_types": ["sha256", "md5"],
                "hatrac_options": {"versioned_urls": True},
                "hatrac_templates": {
                    "hatrac_uri": "/hatrac/{asset_table}/{md5}.{file_name}",
                    "content-disposition": "filename*=UTF-8''{file_name}.{file_ext}",
                },
                "record_query_template": "/entity/{target_table}/MD5={{md5}}&Filename={{file_name}}",
            },
            # {
            #  Upload the records into a  table
            #   "asset_type": "skip",
            ##   "default_columns": ["RID", "RCB", "RMB", "RCT", "RMT"],
            #  "file_pattern": feature_value_regex,  # Sets schema, table,
            #  "ext_pattern": "^.*[.](?P<file_ext>json|csv)$",
            #  "target_table": ["{schema}", "{table}"],
            # },
            {
                #  Upload the records into a  table
                "asset_type": "table",
                "default_columns": ["RID", "RCB", "RMB", "RCT", "RMT"],
                "file_pattern": table_regex,  # Sets schema, table,
                "ext_pattern": "^.*[.](?P<file_ext>json|csv)$",
                "target_table": ["{schema}", "{table}"],
            },
        ],
        "version_update_url": "https://github.com/informatics-isi-edu/deriva-client",
        "version_compatibility": [[">=1.4.0", "<2.0.0"]],
    }


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def upload_directory(
    model: DerivaModel, directory: Path | str
) -> dict[Any, FileUploadState] | None:
    """Upload assets from a directory. This routine assumes that the current upload specification includes a
    configuration for the specified directory.  Every asset in the specified directory is uploaded

    Args:
        model: Model to upload assets to.
        directory: Directory containing the assets and tables to upload.

    Returns:
        Results of the upload operation.

    Raises:
        DerivaMLException: If there is an issue uploading the assets.
    """
    directory = Path(directory)
    if not directory.is_dir():
        raise DerivaMLException("Directory does not exist")

    # Now upload the files by creating an upload spec and then calling the uploader.
    with TemporaryDirectory() as temp_dir:
        spec_file = f"{temp_dir}/config.json"

        with open(spec_file, "w+") as cfile:
            json.dump(bulk_upload_configuration(model), cfile)
        uploader = GenericUploader(
            server={
                "host": model.hostname,
                "protocol": "https",
                "catalog_id": model.catalog.catalog_id,
            },
            config_file=spec_file,
        )
        try:
            uploader.getUpdatedConfig()
            uploader.scanDirectory(directory)
            results = {
                path: FileUploadState(
                    state=UploadState(result["State"]),
                    status=result["Status"],
                    result=result["Result"],
                )
                for path, result in uploader.uploadFiles().items()
            }
        finally:
            uploader.cleanup()
        return results


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def upload_asset(
    model: DerivaModel, file: Path | str, table: Table | str, **kwargs: Any
) -> dict:
    """Upload the specified file into Hatrac and update the associated asset table.

    Args:
        file: path to the file to upload.
        table: Name of the asset table
        model: Model to upload assets to.
        kwargs: Keyword arguments for values of additional columns to be added to the asset table.

    Returns:

    """
    if not model.is_asset(table):
        raise DerivaMLException(f"Table {table} is not an asset table.")

    file_path = Path(file)
    file_name = file_path.name
    file_size = file_path.stat().st_size

    hatrac_path = f"/hatrac/{table.name}/"
    hs = HatracStore(
        "https",
        server=model.catalog.deriva_server.server,
        credentials=model.catalog.deriva_server.credentials,
    )
    md5_hashes = hash_utils.compute_file_hashes(file, ["md5"])["md5"]
    sanitized_filename = urlquote(
        re.sub("[^a-zA-Z0-9_.-]", "_", md5_hashes[0] + "." + file_name)
    )
    hatrac_path = f"{hatrac_path}{sanitized_filename}"

    try:
        # Upload the file to hatrac.
        hatrac_uri = hs.put_obj(
            hatrac_path,
            file,
            md5=md5_hashes[1],
            content_type=mime_utils.guess_content_type(file),
            content_disposition="filename*=UTF-8''" + file_name,
        )
    except Exception as e:
        raise e
    try:
        # Now update the asset table.
        ipath = (
            model.catalog.getPathBuilder().schemas[table.schema.name].tables[table.name]
        )
        return list(
            ipath.insert(
                [
                    {
                        "URL": hatrac_uri,
                        "Filename": file_name,
                        "Length": file_size,
                        "MD5": md5_hashes[0],
                    }
                    | kwargs
                ]
            )
        )[0]
    except Exception as e:
        raise e


class UploadAssetDirectory:
    def __init__(self, model: DerivaModel, schema: str, table: str, prefix: Path):
        self.prefix = prefix
        self.path = asset_table_dir(
            prefix=prefix, asset_schema=schema, asset_table=table
        )
        self.table = table
        self.schema = schema
        self.model = model

    def create_file(self, file_name: str, metadata: dict[str, Any]) -> Path:
        """Return the file in which to place  assets of a specified type are to be uploaded.

        Args:
            file_name: Name of file to which the contents of the asset will be placed
            metadata: Any additional metadata to add to the asset
        Returns:
            Path to directory in which to place assets of type asset_type.
        """
        metadata = metadata or {}
        asset_metadata = self.model.asset_metadata(self.table)
        if not (asset_metadata >= set(metadata.keys())):
            raise DerivaMLException(
                f"Metadata {metadata} does not match asset metadata {asset_metadata}"
            )

        path = self.path
        for m in self.model.asset_metadata(self.table):
            path = path / metadata.get(m, "None")
        path.mkdir(parents=True, exist_ok=True)
        return path / file_name


def test_upload():
    """ """
    ead = execution_asset_dir("foo", "my-rid", "my-asset")
    emd = execution_metadata_dir("foo", "my-rid", "my-metadata")
    _fp = feature_value_path("foo", "my-rid", "my-schema", "my-target", "my-feature")
    fa = feature_asset_dir(
        "foo", "my-rid", "my-schema", "my-target", "my-feature", "my-asset"
    )
    _tp = table_path("foo", "my-schema", "my-table")
    #   _ad = create_asset_dir("foo", "my-schema", "my-asset")
    _is_md = is_execution_metadata_dir(emd)
    _is_ea = is_execution_asset_dir(ead)
    _is_fa = is_feature_asset_dir(fa)



import sys
from deriva.core import ErmrestCatalog, get_credential
import argparse
import os
from pathlib import Path


def update_table_comments(model, schema_name: str, table_name: str, comments_dir: str) -> None:
    table = model.schemas[schema_name].tables[table_name]
    table_comments_dir = Path(comments_dir)/Path(f"{schema_name}/{table_name}")
    for file in os.listdir(table_comments_dir):
        file_path = os.path.join(table_comments_dir, file)
        with open(file_path, "r") as f:
            comment_str = f.read()
            if file.split(".")[0] == table_name:
                table.comment = comment_str
            else:
                table.columns[file.split(".")[0]].comment = comment_str


def update_schema_comments(model, schema_name: str, comments_dir: str) -> None:
    schema_comments_dir = Path(comments_dir)/Path(schema_name)
    for table in os.listdir(schema_comments_dir):
        if not table.endswith(".DS_Store"):
            update_table_comments(model, schema_name, table, comments_dir)


def main():
    scheme = 'https'
    parser = argparse.ArgumentParser()
    parser.add_argument('--hostname', type=str, required=True)
    parser.add_argument('--schema_name', type=str, required=True)
    parser.add_argument('--catalog_id', type=str, required=True)
    parser.add_argument('--comments_dir', type=str, required=True,
                        help="The directory containing the comments files for the whole catalog")
    parser.add_argument('--table_name', type=str,
                        help="Only update the comments for one table")
    args = parser.parse_args()

    credentials = get_credential(args.hostname)
    catalog = ErmrestCatalog(scheme, args.hostname, args.catalog_id, credentials)
    model = catalog.getCatalogModel()
    if args.table_name:
        update_table_comments(model, args.schema_name, args.table_name, args.comments_dir)
        model.apply()
    else:
        update_schema_comments(model, args.schema_name, args.comments_dir)
        model.apply()


if __name__ == '__main__':
    sys.exit(main())



# docs/<schema-name>/<table-name>/[table|<column-name>.Md


import argparse
import sys


def generate_annotation(catalog_id: str, schema: str) -> dict:
    workflow_annotation = {
        "tag:isrd.isi.edu,2016:visible-columns": {
            "*": [
                "RID",
                "Name",
                "Description",
                {
                    "display": {"markdown_pattern": "[{{{URL}}}]({{{URL}}})"},
                    "markdown_name": "URL"
                },
                "Checksum",
                "Version",
                {
                    "source": [
                        {
                            "outbound": [
                                schema,
                                "Workflow_Workflow_Type_fkey"
                            ]
                        },
                        "RID"
                    ]
                }
            ]
        }
    }

    execution_annotation = {
        "tag:isrd.isi.edu,2016:visible-columns": {
            "*": [
                "RID",
                [
                    schema,
                    "Execution_RCB_fkey"
                ],
                "RCT",
                "Description",
                {"source": [
                    {"outbound": [
                        "eye-ai",
                        "Execution_Workflow_fkey"
                    ]
                    },
                    "RID"
                ]
                },
                "Duration",
                "Status",
                "Status_Detail"
            ]
        },
        "tag:isrd.isi.edu,2016:visible-foreign-keys": {
            "detailed": [
                {
                    "source": [
                        {
                            "inbound": [
                                schema,
                                "Dataset_Execution_Execution_fkey"
                            ]
                        },
                        {
                            "outbound": [
                                schema,
                                "Dataset_Execution_Dataset_fkey"
                            ]
                        },
                        "RID"
                    ],
                    "markdown_name": "Dataset"
                },
                {
                    "source": [
                        {
                            "inbound": [
                                schema,
                                "Execution_Assets_Execution_Execution_fkey"
                            ]
                        },
                        {
                            "outbound": [
                                schema,
                                "Execution_Assets_Execution_Execution_Assets_fkey"
                            ]
                        },
                        "RID"
                    ],
                    "markdown_name": "Execution Assets"
                },
                {
                    "source": [
                        {
                            "inbound": [
                                schema,
                                "Execution_Metadata_Execution_fkey"
                            ]
                        },
                        "RID"
                    ],
                    "markdown_name": "Execution Metadata"
                }
            ]
        }
    }

    execution_assets_annotation = {
        "tag:isrd.isi.edu,2016:table-display": {
            "row_name": {
                "row_markdown_pattern": "{{{Filename}}}"
            }
        },
        "tag:isrd.isi.edu,2016:visible-columns": {
            "compact": [
                "RID",
                "URL",
                "Description",
                "Length",
                [
                    schema,
                    "Execution_Assets_Execution_Asset_Type_fkey"
                ],
                # {
                #     "display": {
                #         "template_engine": "handlebars",
                #         "markdown_pattern": "{{#if (eq  _Execution_Asset_Type \"2-5QME\")}}\n ::: iframe []("
                #                             "https://dev.eye-ai.org/~vivi/deriva-webapps/plot/?config=test-line"
                #                             "-plot&Execution_Assets_RID={{{RID}}}){class=chaise-autofill "
                #                             "style=\"min-width: 500px; min-height: 300px;\"} \\n:::\n {{/if}}"
                #     },
                #     "markdown_name": "ROC Plot"
                # }
            ],
            "detailed": [
                "RID",
                "RCT",
                "RMT",
                "RCB",
                "RMB",
                # {
                #     "display": {
                #         "template_engine": "handlebars",
                #         "markdown_pattern": "{{#if (eq _Execution_Asset_Type \"2-5QME\")}} ::: iframe []("
                #                             "https://dev.eye-ai.org/~vivi/deriva-webapps/plot/?config=test-line"
                #                             "-plot&Execution_Assets_RID={{{RID}}}){style=\"min-width:1000px; "
                #                             "min-height:700px; height:70vh;\" class=\"chaise-autofill\"} \\n::: {"
                #                             "{/if}}"
                #     },
                #     "markdown_name": "ROC Plot"
                # },
                "URL",
                "Filename",
                "Description",
                "Length",
                "MD5",
                [
                    schema,
                    "Execution_Assets_Execution_Asset_Type_fkey"
                ]
            ]
        }
    }

    execution_metadata_annotation = {
        "tag:isrd.isi.edu,2016:table-display": {
            "row_name": {
                "row_markdown_pattern": "{{{Filename}}}"
            }
        }
    }

    schema_annotation = {
        "headTitle": "Catalog ML",
        "navbarMenu": {
            "newTab": False,
            "children": [
                {
                    "name": "User Info",
                    "children": [
                        {
                            "url": f"/chaise/recordset/#{catalog_id}/public:ERMrest_Client",
                            "name": "Users"
                        },
                        {
                            "url": f"/chaise/recordset/#{catalog_id}/public:ERMrest_Group",
                            "name": "Groups"
                        },
                        {
                            "url": f"/chaise/recordset/#{catalog_id}/public:ERMrest_RID_Lease",
                            "name": "ERMrest RID Lease"
                        }
                    ]
                },
                {
                    "name": "FaceBase-ML",
                    "children": [
                        {
                            "url": f"/chaise/recordset/#{catalog_id}/{schema}:Workflow",
                            "name": "Workflow"
                        },
                        {
                            "url": f"/chaise/recordset/#{catalog_id}/{schema}:Workflow_Type",
                            "name": "Workflow Type"
                        },
                        {
                            "url": f"/chaise/recordset/#{catalog_id}/{schema}:Execution",
                            "name": "Execution"
                        },
                        {
                            "url": f"/chaise/recordset/#{catalog_id}/{schema}:Execution_Metadata",
                            "name": "Execution Metadata"
                        },
                        {
                            "url": f"/chaise/recordset/#{catalog_id}/{schema}:Execution_Metadata_Type",
                            "name": "Execution Metadata Type"
                        },
                        {
                            "url": f"/chaise/recordset/#{catalog_id}/{schema}:Execution_Assets",
                            "name": "Execution Assets"
                        },
                        {
                            "url": f"/chaise/recordset/#{catalog_id}/{schema}:Execution_Asset_Type",
                            "name": "Execution Asset Type"
                        }
                    ]
                }
            ]
        },
        "navbarBrandText": "ML Data Browser",
        "systemColumnsDisplayEntry": ["RID"],
        "systemColumnsDisplayCompact": ["RID"]
    }

    return {"workflow_annotation": workflow_annotation,
            "execution_annotation": execution_annotation,
            "execution_assets_annotation": execution_assets_annotation,
            "execution_metadata_annotation": execution_metadata_annotation,
            "schema_annotation": schema_annotation
            }


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--catalog_id', type=str, required=True)
    parser.add_argument('--schema_name', type=str, required=True)
    args = parser.parse_args()
    return generate_annotation(args.catalog_id, args.schema_name)


if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations

import json
from typing import Optional

from pydantic import (
    BaseModel,
    conlist,
    ConfigDict,
)
from pathlib import Path


from .dataset_aux_classes import DatasetSpec
from .deriva_definitions import RID


class Workflow(BaseModel):
    """A specification of a workflow.  Must have a name, URI to the workflow instance, and a type.  The workflow type
    needs to be an existing-controlled vocabulary term.

    Attributes:
        name: The name of the workflow
        url: The URI to the workflow instance.  In most cases should be a GitHub URI to the code being executed.
        workflow_type: The type of the workflow.  Must be an existing controlled vocabulary term.
        version: The version of the workflow instance.  Should follow semantic versioning.
        description: A description of the workflow instance.  Can be in Markdown format.
    """

    name: str
    url: str
    workflow_type: str
    version: Optional[str] = None
    description: str = None


class ExecutionConfiguration(BaseModel):
    """Define the parameters that are used to configure a specific execution.

    Attributes:
        datasets: List of dataset_table RIDS, MINIDS for datasets to be downloaded prior to execution.  By default,
                     all  the datasets are materialized. However, if the assets associated with a dataset_table are not
                     needed, a dictionary that defines the rid and the materialization parameter for the
                     download_dataset_bag method can be specified, e.g.  datasets=[{'rid': RID, 'materialize': True}].
        assets: List of assets to be downloaded prior to execution.  The values must be RIDs in an asset table
        workflow: A workflow instance.  Must have a name, URI to the workflow instance, and a type.
        description: A description of the execution.  Can use Markdown format.
    """

    datasets: conlist(DatasetSpec) = []
    assets: list[RID | str] = []  # List of RIDs to model files.
    workflow: Workflow
    description: str = ""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @staticmethod
    def load_configuration(path: Path) -> ExecutionConfiguration:
        """Create a ExecutionConfiguration from a JSON configuration file.

        Args:
          path: File containing JSON version of execution configuration.

        Returns:
          An execution configuration whose values are loaded from the given file.
        """
        with open(path) as fd:
            config = json.load(fd)
        return ExecutionConfiguration.model_validate(config)

    def download_execution_configuration(
        self, configuration_rid: RID
    ) -> ExecutionConfiguration:
        """Create an ExecutionConfiguration object from a catalog RID that points to a JSON representation of that
        configuration in hatrac

        Args:
            configuration_rid: RID that should be to an asset table that refers to an execution configuration

        Returns:
            A ExecutionConfiguration object for configured by the parameters in the configuration file.
        """
        AssertionError("Not Implemented")
        return ExecutionConfiguration.load_configuration(configuration_rid)

        # configuration = self.retrieve_rid(configuration_rid)
        # with NamedTemporaryFile("w+", delete=False, suffix=".json") as dest_file:
        #    hs = HatracStore("https", self.host_name, self.credential)
        #    hs.get_obj(path=configuration["URL"], destfilename=dest_file.name)
        #    return ExecutionConfiguration.load_configuration(Path(dest_file.name))
